#include <string.h>
#include <stdio.h>
#include <sys/exception.h>
#include <services/services.h>
#include <drivers/adi_dev.h>
#include <ADI_ETHER_BCEA.h>
#include <fpga_irq.h>

#include "bcea_hw.h"

typedef struct ADI_ETHER_BCEA_DATA {
	u32 Open;
	u32 Running;
	u32 Cache;
	u8 macAddr[6];

	ADI_ETHER_BUFFER *rxBuf, **rxBufTailPtr;
	ADI_ETHER_BUFFER *txBuf, **txBufTailPtr;

	void *CriticalData;
	ADI_DEV_DEVICE_HANDLE DeviceHandle;
	ADI_DCB_HANDLE DCBHandle;
	ADI_DCB_CALLBACK_FN DMCallback;
	ADI_DEV_DIRECTION Direction;
} ADI_ETHER_BCEA_DATA;
static ADI_ETHER_BCEA_DATA EtherDev = {
	.Open = 0,
	.Running = 0,
	.Cache = 1
};

static void tx_irq_handler(u8 irq, void *ptr)
{
	unsigned addr;
	ADI_ETHER_BCEA_DATA *dev = ptr;
	ADI_ETHER_BUFFER *iter, *acked;
	ADI_DCB_RESULT result;

	while((addr = *pBCEA_TX_FREE) != 0xFFFFFFFF) {
		/* nothing to do so far */
		for(iter=dev->txBuf; iter; iter=iter->pNext)
			if(((u32)(iter->Data))+2 == addr)
				iter->ProcessedFlag = 1;

		*pBCEA_RX_POLL;
	}

	acked = NULL;
	for(iter=dev->txBuf; iter; iter=iter->pNext) {
		if(iter->pNext && !iter->pNext->ProcessedFlag)
			acked = iter;
		if(!iter->ProcessedFlag)
			break;
	}

	if(iter == dev->txBuf)
		return;	

	if(acked) {
		acked->pNext = NULL;
		acked = dev->txBuf;
		dev->txBuf = iter;
	} else {
		acked = dev->txBuf;
		dev->txBuf = NULL;
		dev->txBufTailPtr = &(dev->txBuf);
	}

	if(dev->DCBHandle)
		result = adi_dcb_Post(dev->DCBHandle, 9, dev->DMCallback, dev->DeviceHandle,
							  ADI_ETHER_EVENT_FRAME_XMIT, acked->CallbackParameter);
	else
		(dev->DMCallback)(dev->DeviceHandle, ADI_ETHER_EVENT_FRAME_XMIT, acked->CallbackParameter);
}

static void rx_irq_handler(u8 irq, void *ptr)
{
	unsigned len;
	ADI_ETHER_BCEA_DATA *dev = ptr;
	ADI_ETHER_BUFFER *pkt, *chain = NULL, **chain_tail;
	ADI_DCB_RESULT result;
	
	chain_tail = &chain;
	
	while(len = *pBCEA_RX_POLL) {
		pkt = dev->rxBuf;

		if(pkt) {
			pkt->ProcessedElementCount = len + 6;
			if(pkt->ProcessedElementCount <= pkt->ElementCount) {
				*(u16 *)(pkt->Data) = len;
				memcpy(((u8 *)(pkt->Data))+2, (void *)(pBCEA_RX_DATA(0)), len);

				pkt->ProcessedFlag = 1;
				pkt->StatusWord = 0x3000 + len;
				
				dev->rxBuf = pkt->pNext;
				if(!dev->rxBuf)
					dev->rxBufTailPtr  = &(dev->rxBuf);

				pkt->pNext = NULL;
				*chain_tail = pkt;
				chain_tail = &(pkt->pNext);
			}
		}

		*pBCEA_RX_ACK = 0;
		*pBCEA_RX_POLL;
	}

	if(chain) {
		if(dev->DCBHandle)
			result = adi_dcb_Post(dev->DCBHandle, 9, dev->DMCallback, dev->DeviceHandle,
								  ADI_ETHER_EVENT_FRAME_RCVD, chain->CallbackParameter);
		else
			(dev->DMCallback)(dev->DeviceHandle, ADI_ETHER_EVENT_FRAME_RCVD, chain->CallbackParameter);
	}
}

/*********************************************************************
*
*   Function:       pddOpen
*
*   Description:    Opens the Network Driver and does initialization.
*
*********************************************************************/

static u32 adi_pdd_Open(                // Open a device
    ADI_DEV_MANAGER_HANDLE  ManagerHandle,  // device manager handle
    u32                     DeviceNumber,   // device number
    ADI_DEV_DEVICE_HANDLE   DeviceHandle,   // device handle
    ADI_DEV_PDD_HANDLE      *pPDDHandle,    // pointer to PDD handle location
    ADI_DEV_DIRECTION       Direction,      // data direction
    void            *pCriticalRegionArg,        // critical region imask storage location
    ADI_DMA_MANAGER_HANDLE  DMAHandle,      // handle to the DMA manager
    ADI_DCB_HANDLE          DCBHandle,      // callback handle
    ADI_DCB_CALLBACK_FN     DMCallback      // device manager callback function
)
{
    u32             Result;     // return value
    ADI_ETHER_BCEA_DATA *dev= &EtherDev;

    Result = ADI_DEV_RESULT_DEVICE_IN_USE;

    if(!dev->Open)
    {
		*pEBIU_AMGCTL = 0x0002; /* Enable Bank 0, 32-bit bus */
		*pEBIU_AMBCTL0 = 0xFFC22416; /* wait times */

		if((*pBCEA_RX_LINK >> 16) != 0xBCEA) {
			printf("BCEA: No hardware found.\n");
			Result = ADI_DEV_RESULT_FAILED;
			return (Result);
		}

		printf("BCEA: Hardware v%d.%d found.\n", (*pBCEA_RX_LINK >> 8) & 0xFF, (*pBCEA_RX_LINK >> 4) & 0xF);

		while(*pBCEA_RX_POLL != 0x00000000) *pBCEA_RX_ACK = 0;
		while(*pBCEA_TX_FREE != 0xFFFFFFFF) ;

		firq_disable(FIRQ_BCEA_TX);
		firq_disable(FIRQ_BCEA_RX);
		firq_disable(FIRQ_BCEA_LINK);

		firq_set_handler(FIRQ_BCEA_TX, tx_irq_handler, dev);
		firq_set_handler(FIRQ_BCEA_RX, rx_irq_handler, dev);
										    	
    	dev->CriticalData = pCriticalRegionArg;
        dev->DeviceHandle = DeviceHandle;
        dev->DCBHandle = DCBHandle;
        dev->DMCallback = DMCallback;
        dev->Direction = Direction;

        dev->rxBuf = NULL;
        dev->rxBufTailPtr = &(dev->rxBuf);

        dev->txBuf = NULL;
        dev->txBufTailPtr = &(dev->txBuf);
                
        dev->Open = 1;
        Result = ADI_DEV_RESULT_SUCCESS;
    }

    if (Result != ADI_DEV_RESULT_SUCCESS) return (Result);

    *pPDDHandle = (ADI_DEV_PDD_HANDLE *)&EtherDev;
    return(ADI_DEV_RESULT_SUCCESS);
}


/*********************************************************************
*
*   Function:       pddRead
*
*   Description:    Gives list of read buffers to the driver
*
*********************************************************************/
static u32 adi_pdd_Read(        // Reads data or queues an inbound buffer to a device
    ADI_DEV_PDD_HANDLE PDDHandle,   // PDD handle
    ADI_DEV_BUFFER_TYPE BufferType, // buffer type
    ADI_DEV_BUFFER *pBuffer         // pointer to buffer
)
{
    u32         Result;             // return value
    ADI_ETHER_BCEA_DATA *dev = (ADI_ETHER_BCEA_DATA *)PDDHandle;
    ADI_ETHER_BUFFER *buf = (ADI_ETHER_BUFFER*)pBuffer, *iter;

    if(!buf)
    	return(ADI_DEV_RESULT_SUCCESS);
    
	*dev->rxBufTailPtr = buf;
	for(iter = buf; iter->pNext; iter=iter->pNext) ;
	dev->rxBufTailPtr = &(iter->pNext);

	return(ADI_DEV_RESULT_SUCCESS);
}


/*********************************************************************
*
*   Function:       pddWrite
*
*   Description:    Sends packet over the physical channel
*
*********************************************************************/
static u32 adi_pdd_Write(           // Writes data or queues an outbound buffer to a device
    ADI_DEV_PDD_HANDLE PDDHandle,               // PDD handle
    ADI_DEV_BUFFER_TYPE BufferType,             // buffer type
    ADI_DEV_BUFFER *pBuffer                     // pointer to buffer
)
{
    u32         Result;             // return value
    ADI_ETHER_BCEA_DATA *dev = (ADI_ETHER_BCEA_DATA *)PDDHandle;
    ADI_ETHER_BUFFER *buf = (ADI_ETHER_BUFFER*)pBuffer, *iter;
    u16 len;
    u32 *dataBuf;
   
    for(iter=buf; iter; iter=iter->pNext) {
    	len = *(u16 *)(iter->Data);
    	dataBuf = (u32 *)(((u8 *)iter->Data) + 2);
  
    	if((u32)dataBuf & 3) {
    		printf("Invalid buffer alignment: %p\n", iter->Data);
    		iter->ProcessedFlag = 1;
    		continue;
    	} else
    		iter->ProcessedFlag = 0;

#ifdef TX_DEBUG    		
    	printf("BCEA: Transmit %08X:%02X [ %08X %08X %08X %08X\n", (u32)dataBuf, len,
    		dataBuf[0], dataBuf[1], dataBuf[2], dataBuf[3]);
    	printf("   %08X %08X %08X %08X %08X %08X %08X %08X... ]\n",
    		dataBuf[4], dataBuf[5], dataBuf[6], dataBuf[7],
    		dataBuf[8], dataBuf[9], dataBuf[10], dataBuf[11]);
#endif
    	*pBCEA_TX_START = (u32)dataBuf;
    	if(len < 64)
    		len = 64;
    	*pBCEA_TX_LEN = len | BCEA_TX_LAST;
    	*pBCEA_TX_COMMIT = 0;
    }

    *(dev->txBufTailPtr) = buf;
    
    return(ADI_DEV_RESULT_SUCCESS);
}

/*********************************************************************
*
*   Function:       pddControl
*
*   Description:    List of I/O control commands to the driver
*
*********************************************************************/
static u32 adi_pdd_Control(         // Sets or senses a device specific parameter
    ADI_DEV_PDD_HANDLE PDDHandle,               // PDD handle
    u32 Command,                    // command ID
    void *pArg                              // pointer to argument
)
{
    u32         Result;             // return value
    ADI_ETHER_BCEA_DATA *dev = (ADI_ETHER_BCEA_DATA *)PDDHandle;
    ADI_ETHER_MEM_SIZES *msizes;
    ADI_ETHER_SUPPLY_MEM *memsup;
    int maxbuf,i;
    u32 *prefix;
    ADI_ETHER_BUFFER_COUNTS *bufcnts;
    u32 MultiCastGroupAddr;

    // avoid casts
    Result = ADI_DEV_RESULT_SUCCESS;
    
    // CASEOF (Command ID)
    switch (Command) {
        case ADI_DEV_CMD_SET_DATAFLOW_METHOD:
        	/* ignore call */
            break;
        case ADI_DEV_CMD_SET_DATAFLOW:
        	/* ignore call */
            break;
            
        case ADI_DEV_CMD_GET_PERIPHERAL_DMA_SUPPORT:
            (*(int *)pArg) = false;
            break;

        case ADI_ETHER_CMD_MEM_SIZES:
            msizes = (ADI_ETHER_MEM_SIZES *)pArg;
            msizes->BaseMemSize = 0;
            msizes->MemPerRecv = 0;
            msizes->MemPerXmit = 0;
            break;

        case ADI_ETHER_CMD_SUPPLY_MEM:
            memsup = (ADI_ETHER_SUPPLY_MEM *)pArg;
            break;

        case ADI_ETHER_CMD_GET_MAC_ADDR:
        	memcpy(pArg, dev->macAddr, 6);
            break;

        case ADI_ETHER_CMD_SET_MAC_ADDR:
            memcpy(dev->macAddr,(unsigned char*)pArg,6);
        	printf("BCEA: MAC addr is %02X:%02X:%02X:%02X:%02X:%02X\n",
        		dev->macAddr[0], dev->macAddr[1], dev->macAddr[2], dev->macAddr[3], dev->macAddr[4], dev->macAddr[5]);
			*pBCEA_RX_MACL = dev->macAddr[0] | ((u32)(dev->macAddr[1]) << 8) |
			                 ((u32)(dev->macAddr[2]) << 16) | ((u32)(dev->macAddr[3]) << 24);
			*pBCEA_RX_MACH = dev->macAddr[4] | ((u32)(dev->macAddr[5]) << 8) |
							 (dev->Running ? BCEA_RXEN : 0);	
            break;

        case ADI_ETHER_CMD_GET_BUFFER_PREFIX:
            prefix = (u32 *)pArg;
            *prefix = 2;
            break;

        case ADI_ETHER_CMD_UNPROCESSED_BUFFER_COUNTS:
            bufcnts = (ADI_ETHER_BUFFER_COUNTS *)pArg;
            bufcnts->RcvrBufferCnt = 0;
            bufcnts->XmitBufferCnt = 0;
            break;

        case ADI_ETHER_CMD_GET_MIN_RECV_BUFSIZE:
            prefix = (u32 *)pArg;
            *prefix = 1500; /* for now */
            break;
/*
        case ADI_ETHER_CMD_GET_PHY_REGS:
                short *arg = (short *)pArg;
                int ic;

                for (ic=0;ic<=PHY_MASK_REG;ic++) arg[ic] = LAN91C111_read_phy_register(ic);
            break;
*/
        case ADI_ETHER_CMD_BUFFERS_IN_CACHE:
            i = (int)(pArg);
            dev->Cache = (i != 0);
            break;

		case ADI_ETHER_CMD_START:
			dev->Running = 1;
			*pBCEA_RX_MACH = dev->macAddr[4] | ((u32)(dev->macAddr[5]) << 8) | BCEA_RXEN;
			firq_enable(FIRQ_BCEA_TX);
			firq_enable(FIRQ_BCEA_RX);
			printf("BCEA: Started.\n");
            break;

        case ADI_ETHER_ADD_MULTICAST_MAC_FILTER:
            MultiCastGroupAddr = (u32)pArg;
//            printf("BCEA: AddMCast(0x%08X)\n", MultiCastGroupAddr);
			break;

		case ADI_DEV_CMD_TABLE:
            break;

        case ADI_DEV_CMD_END:
            break;

        case ADI_ETHER_CMD_SHUTDOWN_DRIVER:
//            ShutDownDriver(PDDHandle);
            break;

        default:
            // we don't understand this command
            printf("Unsupported command: 0x%08X\n", Command);
            Result = ADI_DEV_RESULT_NOT_SUPPORTED;

    }

    // return
    return(Result);
}

/*********************************************************************
*
*   Function:       pddClose
*
*   Description:    Closes the driver and releases any memory
*
*********************************************************************/

static u32 adi_pdd_Close(       // Closes a device
    ADI_DEV_PDD_HANDLE PDDHandle    // PDD handle
)
{
    u32         Result = ADI_DEV_RESULT_SUCCESS;                // return value
    ADI_ETHER_BCEA_DATA *dev = (ADI_ETHER_BCEA_DATA *)PDDHandle;

    printf("BCEA:Closing.\n");
    
    dev->Open = 0;
    dev->Running = 0;
    return(Result);
}

/**************************************************************************
 *
 * BCEA entry point
 *
 **************************************************************************/
ADI_DEV_PDD_ENTRY_POINT ADI_ETHER_BCEA_Entrypoint  = {
    adi_pdd_Open,
    adi_pdd_Close,
    adi_pdd_Read,
    adi_pdd_Write,
    adi_pdd_Control
};

