#include <stdio.h>
#include <fpga_spi.h>

#define FPGA_SPIF_BASE   0x20000000
#define SPIF_WRITE8		 0x6000
#define SPIF_WRITE16	 0x6004
#define SPIF_WRITE24     0x6008
#define SPIF_WRITE32	 0x600c
#define SPIF_CTL		 0x6010
#define SPIF_MISO		 0x6014

#define SPIF_CTL_ENABLE		1
#define SPIF_CTL_DISABLE	0

#define SPIF_PTR(x)		((volatile unsigned int *)(FPGA_SPIF_BASE + (x)))

#define pSPIF_WRITE(x)	SPIF_PTR(SPIF_WRITE8-4+((x)*4))
#define pSPIF_WRITE8	SPIF_PTR(SPIF_WRITE8)
#define pSPIF_WRITE16	SPIF_PTR(SPIF_WRITE16)
#define pSPIF_WRITE24	SPIF_PTR(SPIF_WRITE24)
#define pSPIF_WRITE32	SPIF_PTR(SPIF_WRITE32)
#define pSPIF_CTL		SPIF_PTR(SPIF_CTL)
#define pSPIF_MISO		SPIF_PTR(SPIF_MISO)

static unsigned fpga_spi_trans(unsigned data, unsigned len)
{
	*pSPIF_CTL = SPIF_CTL_ENABLE;
	*pSPIF_WRITE(len) = data;
	*pSPIF_CTL;
	while (*pSPIF_CTL);
	return *pSPIF_WRITE8;
}

static inline void fpga_spi_done(void)
{
	*pSPIF_CTL = SPIF_CTL_DISABLE;
}

static unsigned fpga_spi_flash_query(void)
{
	unsigned id;

	/* WRDI */
	fpga_spi_trans(0x04, 1);
	fpga_spi_done();

	/* DBSY */
	fpga_spi_trans(0x80, 1);
	fpga_spi_done();

	fpga_spi_trans(0x90000000, 4);
	id = fpga_spi_trans(0x0000, 2);
	fpga_spi_done();
	
	printf("FPGA Flash ID = 0x%04X\n", id);
		
	return 0;
}

static void fpga_spi_wait_bsy(void)
{
	*pSPIF_CTL = SPIF_CTL_ENABLE;
	*pSPIF_MISO;
	while(!*pSPIF_MISO);
	*pSPIF_CTL = SPIF_CTL_DISABLE;
}

static void fpga_spi_unprotect(void)
{
	/* EWSR */
	fpga_spi_trans(0x50, 1);
	fpga_spi_done();

	/* WRSR */
	fpga_spi_trans(0x0100, 2);
	fpga_spi_done();
}

static int fpga_spi_chip_erase(void)
{
	/* WREN */
	fpga_spi_trans(0x06, 1);
	fpga_spi_done();

	/* CHIP ERASE */
	fpga_spi_trans(0x60, 1);
	fpga_spi_done();

	/* RDSR */
	fpga_spi_trans(0x05, 1);
	while(fpga_spi_trans(0x00, 1) & 1) ;
	fpga_spi_done();

	return 0;	
}

static int fpga_spi_write(unsigned base, unsigned len, unsigned short *data)
{
	unsigned i;

	/* EBSY */
	fpga_spi_trans(0x70, 1);
	fpga_spi_done();

	/* WREN */
	fpga_spi_trans(0x06, 1);
	fpga_spi_done();

	/* AAI */
	fpga_spi_trans(0xAD000000 | base, 4);
	fpga_spi_trans((data[0] >> 8) | (data[0] << 8), 2);
	fpga_spi_done();

	fpga_spi_wait_bsy();

	for(i=1; i<len; i++) {
		fpga_spi_trans(0xAD0000 | (data[i] >> 8) | ((data[i] & 0xFF) << 8), 3);
		fpga_spi_done();

		fpga_spi_wait_bsy();
	}

	/* WRDI */
	fpga_spi_trans(0x04, 1);
	fpga_spi_done();

	/* DBSY */
	fpga_spi_trans(0x80, 1);
	fpga_spi_done();

	return 0;	
}

static int fpga_spi_verify(unsigned base, unsigned len, unsigned char *data)
{
	unsigned i, err = 0;
	unsigned char rd;

	/* FAST READ */
	fpga_spi_trans(0x0B000000 | base, 4);
	fpga_spi_trans(0x00, 1);
	for(i=0; i<len; i++) {
		rd = fpga_spi_trans(0x00, 1);
		if(rd != data[i]) {
			printf("FPGA SPI Flash: at %06X should be %02X, is %02X\n", base + i, data[i], rd);
			err ++;
			if(err > 10) {
				fpga_spi_done();
				return err;
			}
		}
	}
	fpga_spi_done();

	return err;	
}
