#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "micser.h"

#define CS_IDLE      0
#define CS_WAIT_REQ  1
#define CS_TX_HDR    2
#define CS_TX_DATA   3
#define CS_TX_ERR    4

typedef struct {
    int state;
    int fd;
    u32 ipaddr;
    char buf[MS_MAX_REQ_SIZE];
    u32 ptr, len;
    ms_file *file;
    ms_gen_func link;
    u32 dlen, doff, dgen;
    u8 *dptr;
} conn_state;

static conn_state conns[MS_MAX_CONNS];
static int num_conns;
static u32 alloc_conn(void)
{
    u32 i;
    for(i=0; i<MS_MAX_CONNS; i++)
	if(!conns[i].state)
	    break;

    if(i >= MS_MAX_CONNS)
	return 0xFFFFFFFF;

    conns[i].dgen = 0;
    conns[i].ptr = 0;
    conns[i].state = CS_WAIT_REQ;
    num_conns ++;
    return i;
}
static void free_conn(u32 conn)
{
    if(conns[conn].dgen && conns[conn].dptr)
	free(conns[conn].dptr);
    num_conns --;
    conns[conn].state = CS_IDLE;
}
static int scan_conn(fd_set *rd_set, fd_set *wr_set)
{
    u32 i, max=0;

    FD_ZERO(rd_set);
    FD_ZERO(wr_set);

    for(i=0; i<MS_MAX_CONNS; i++) {
	if(conns[i].state == CS_WAIT_REQ)
	    FD_SET(conns[i].fd, rd_set);
	if(conns[i].state == CS_TX_HDR || conns[i].state == CS_TX_ERR || conns[i].state == CS_TX_DATA)
	    FD_SET(conns[i].fd, wr_set);
	if(conns[i].state && conns[i].fd > max)
	    max = conns[i].fd;
    }

    return max;
}

static int parse_fname(u32 conn, char *fname, void *root)
{
    ms_file *iter = root;
    int ent;
    char *p, *q;

    if(fname[0] != '/')
	return 1;

    p = fname + 1;
    while(*p) {
	q = strchr(p, '/');
	if(q)
	    *q = 0;

	if(!iter)
	    return 1;

	if(iter->type != MS_FT_DIRECTORY)
	    return 1;
	for(ent=0; ent<iter->dir.count; ent++)
	    if(!strcmp(p, (char *)(iter->dir.entries[ent].name)))
		break;
	if(ent >= iter->dir.count)
	    return 1;
	iter = (ms_file *)(iter->dir.entries[ent].ptr + (u8 *)root);

	if(!q)
	    break;
	p = q+1;
    }

    if(!iter)
	return 1;

    if(iter->type == MS_FT_DIRECTORY) {
	if(iter->dir.defent == 0xFFFFFFFF)
	    return 1;
	iter = (ms_file *)(iter->dir.entries[iter->dir.defent].ptr + (u8 *)root);
    }

    if(iter->type == MS_FT_DIRECTORY)
	return 1;

    conns[conn].file = iter;

    return 0;
}

static int process_file(u32 conn, ms_gen_func *link)
{
    u32 result;

    conns[conn].dptr = NULL;
    conns[conn].link = NULL;
    conns[conn].dgen = 0;

    if(conns[conn].file->type == MS_FT_STATIC) {
	conns[conn].dlen = conns[conn].file->st.size;
	conns[conn].dptr = conns[conn].file->st.data;
	return 0;
    }

    if(conns[conn].file->type == MS_FT_GENERATED &&
       !conns[conn].file->gen.partial) {
        result = link[conns[conn].file->gen.link](0, 0, &(conns[conn].dlen), &(conns[conn].dptr));
	conns[conn].dgen = !result;
	return result;
    }

    if(conns[conn].file->type == MS_FT_GENERATED) {
	conns[conn].link = link[conns[conn].file->gen.link];
	return conns[conn].link(0, 0, &(conns[conn].dlen), NULL);
    }

    return 1;
}

static void start_data(u32 conn)
{
    u8 *buf;

    if(conns[conn].link) {
	buf = (u8 *)conns[conn].buf;
	conns[conn].len = MS_MAX_REQ_SIZE;
	conns[conn].link(1, 0, &(conns[conn].len), &buf);
	conns[conn].doff = conns[conn].len;
    } else
	conns[conn].len = conns[conn].dlen;

    conns[conn].ptr = 0;
    conns[conn].state = CS_TX_DATA;
}

static void get_request(u32 conn, void *root, ms_gen_func *link)
{
    int count, total = 0, fullhdr = 0;
    u32 i, st;
    char *p, *fname, *mime;

    while((count = read(conns[conn].fd, conns[conn].buf+conns[conn].ptr, MS_MAX_REQ_SIZE-1-conns[conn].ptr)) > 0) {
	conns[conn].ptr += count;
	total += count;
    }

    if((errno && errno != EWOULDBLOCK) || !total) {
	close(conns[conn].fd);
	free_conn(conn);
	return;
    }

    if(conns[conn].ptr < 2)
	return;

    /* test for double newline */
    st = 0;
    for(i=0; i<conns[conn].ptr; i++)
	if(conns[conn].buf[i] == '\n') {
	    st ++;
	    if(st > 1)
		break;
	} else if(conns[conn].buf[i] != '\r')
	    st = 0;
    if(st < 2)
	return;

    conns[conn].buf[conns[conn].ptr] = 0;

    fname = strchr(conns[conn].buf, ' ');
    if(!fname)
	goto invalid_req;
    p = strpbrk(fname+1, " \n\r");
    *(fname++) = 0;
    if(!p)
	goto invalid_req;
    if(*p == ' ' && !strncmp(p+1, "HTTP/1.", 7))
	fullhdr = 1;
    *(p++) = 0;

    /* only support GET for now */
    if(strcmp(conns[conn].buf, "GET"))
	goto invalid_req;

    if(parse_fname(conn, fname, root)) {
	strcpy(conns[conn].buf, "HTTP/1.0 404 File Not Found\r\n\r\n<h1>404 File Not Found</h1>");
	goto other_error;
    }

    if(process_file(conn, link)) {
	strcpy(conns[conn].buf, "HTTP/1.0 503 Service Unavailable\r\n\r\n<h1>503 Service Unavailable</h1>");
	goto other_error;
    }

    if(fullhdr) {
	mime = "text/plain";
	if(conns[conn].file->type == MS_FT_STATIC)
	    mime = ((char *)root) + conns[conn].file->st.mime;
	else if(conns[conn].file->type == MS_FT_GENERATED)
	    mime = ((char *)root) + conns[conn].file->gen.mime;

	if(!conns[conn].link || conns[conn].dlen)
	    sprintf(conns[conn].buf, "HTTP/1.1 200 OK\r\nContent-Length: %d\r\nConnection: close\r\nContent-Type: %s\r\n\r\n", conns[conn].dlen, mime);
	else
	    sprintf(conns[conn].buf, "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Type: %s\r\n\r\n", mime);

	conns[conn].ptr = 0;
	conns[conn].len = strlen(conns[conn].buf);
	conns[conn].state = CS_TX_HDR;
    } else
	start_data(conn);
    return;

invalid_req:
    strcpy(conns[conn].buf, "HTTP/1.0 400 Bad Request\r\n\r\n<h1>400 Bad Request</h1>");
other_error:
    conns[conn].ptr = 0;
    conns[conn].len = strlen(conns[conn].buf);
    conns[conn].state = CS_TX_ERR;
}

static void send_headers(u32 conn)
{
    int count, total = 0;

    while((count = write(conns[conn].fd, conns[conn].buf+conns[conn].ptr, conns[conn].len-conns[conn].ptr)) > 0) {
	conns[conn].ptr += count;
	total += count;
    }

    if((errno && errno != EWOULDBLOCK) || !total) {
	close(conns[conn].fd);
	free_conn(conn);
	return;
    }

    if(conns[conn].ptr != conns[conn].len)
	return;

    if(conns[conn].state == CS_TX_ERR || (!conns[conn].dlen && !conns[conn].link)) {
	close(conns[conn].fd);
	free_conn(conn);
	return;
    }

    start_data(conn);
}

static void send_data(u32 conn)
{
    int count, total = 0;
    u8 *buf;

    if(conns[conn].link)
	buf = (u8 *)conns[conn].buf;
    else
	buf = conns[conn].dptr;

    while((count = write(conns[conn].fd, buf+conns[conn].ptr, conns[conn].len-conns[conn].ptr)) > 0) {
	conns[conn].ptr += count;
	total += count;
    }

    if((errno && errno != EWOULDBLOCK) || !total) {
	close(conns[conn].fd);
	free_conn(conn);
	return;
    }

    if(conns[conn].ptr != conns[conn].len)
	return;

    if(conns[conn].link) {
	conns[conn].len = MS_MAX_REQ_SIZE;
	conns[conn].link(1, conns[conn].doff, &(conns[conn].len), &buf);
	conns[conn].doff += conns[conn].len;
	if(conns[conn].len)
	    return;
    }

    close(conns[conn].fd);
    free_conn(conn);
}

void start_server(void *root, ms_gen_func *link, u16 port)
{
    int listen_socket;
    struct sockaddr_in server_addr, client_addr;
    socklen_t client_addr_size;
    int maxfd;
    fd_set rd_set, wr_set;
    u32 conn;

    memset(&conns, 0, sizeof(conns));
    num_conns = 0;

    listen_socket = socket(AF_INET, SOCK_STREAM, 0);
    if(listen_socket < 0) {
	printf("micser: socket() failed\n");
	return;
    }

    memset(&server_addr, 0, sizeof(server_addr));
    server_addr.sin_family = AF_INET;
    server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    server_addr.sin_port = htons(port);

    if(bind(listen_socket, (struct sockaddr *)&server_addr, sizeof(server_addr)) < 0) {
	printf("micser: bind() failed\n");
	return;
    }

    if(listen(listen_socket, MS_MAX_CONNS)) {
	printf("micser: listen() failed\n");
	return;
    }

    while(1) {
	maxfd = scan_conn(&rd_set, &wr_set);

	if(num_conns < MS_MAX_CONNS) {
	    FD_SET(listen_socket, &rd_set);
	    if(listen_socket > maxfd)
		maxfd = listen_socket;
	}

	if(select(maxfd+1, &rd_set, &wr_set, NULL, NULL) == -1) {
	    printf("micser: select() failed\n");
	    return;
	}

	for(conn=0; conn<MS_MAX_CONNS; conn++) {
	    if(conns[conn].state == CS_WAIT_REQ && FD_ISSET(conns[conn].fd, &rd_set))
		get_request(conn, root, link);
	    if((conns[conn].state == CS_TX_HDR || conns[conn].state == CS_TX_ERR) && FD_ISSET(conns[conn].fd, &wr_set))
		send_headers(conn);
	    if(conns[conn].state == CS_TX_DATA && FD_ISSET(conns[conn].fd, &wr_set))
		send_data(conn);
	}

	if(FD_ISSET(listen_socket, &rd_set)) {
	    conn = alloc_conn();
	    if(conn == 0xFFFFFFFF) {
		printf("micser: alloc_conn() failed\n");
		return;
	    }

	    client_addr_size = sizeof(client_addr);
	    conns[conn].fd = accept(listen_socket, (struct sockaddr *)&client_addr, &client_addr_size);
	    if(conns[conn].fd != -1) {
		if(fcntl(conns[conn].fd, F_SETFL, O_NONBLOCK) == -1) {
		    printf("micser: fcntl() failed\n");
		    close(conns[conn].fd);
		    free_conn(conn);
		}
	    } else {
		printf("micser: accept() failed\n");
		free_conn(conn);
	    }
	}
    }
}
