// Timing Sequence Controller

module tsc(clk, reset, done, sync, power, pc,
    t_select, t_clk_yl, t_reset, t_syl, t_sync_yl, t_sync_yr, t_clk_y, t_syr, t_sync_y, t_clk_yr, t_sync_x, t_clk_x, t_pc, t_shr, t_shs,
    t_mclk, t_rsmp, t_vsmp,
    t_ppiclk, t_ppihs, t_ppivs,
    wclk, we, wa, wd, wipe, wip);

input clk;
input reset;
output done;
input sync;
input power;
output [8:0] pc;

output t_select;
output t_clk_yl;
output t_reset;
output t_syl;
output t_sync_yl;
output t_sync_yr;
output t_clk_y;
output t_syr;
output t_sync_y;
output t_clk_yr;
output t_sync_x;
output t_clk_x;
output t_pc;
output t_shr;
output t_shs;

output t_mclk;
output t_rsmp;
output t_vsmp;

output t_ppiclk;
output t_ppihs;
output t_ppivs;

input wclk;
input we;
input [8:0] wa;
input [31:0] wd;

input wipe;
input [8:0] wip;

// generate mclk and oclk

wire ppi_clk_gate;
reg mclk = 1'b0;
reg oclk = 1'b0;
wire p_ppiclk = oclk & ppi_clk_gate;
wire p_mclk = mclk;
always @(posedge clk) begin
    oclk <= !oclk;
    if(oclk)
	mclk <= !mclk;
end

// move wipe/wip to clk domain

reg nipe;
reg [8:0] nip;
reg wipetog, nipetog, nipetog_q, nipetog_qq;
always @(posedge wclk) begin
    if(wipe) begin
	nip <= wip;
        wipetog <= ~wipetog;
    end

    if(reset)
	wipetog <= 1'b0;
end
always @(posedge clk) begin
    nipetog <= wipetog;
    nipetog_q <= nipetog;
    nipetog_qq <= nipetog_q;
    nipe <= nipetog_qq ^ nipetog_q;
end

// advance program counter & fetch

reg [8:0] pc = 9'd0;
wire [29:0] mr;
reg [29:0] mr_q, mr_s;
wire adv;
reg adv_q;
wire [29:0] rr = adv_q ? mr : mr_s;
reg jump_q, nipe_q;
wire jump = (adv && rr[29:28] == 2'b10) && !jump_q && !nipe_q;
wire loop;
wire [8:0] looppc;
wire [8:0] npc = loop ? looppc : jump ? rr[24:16] : pc+1;
always @(posedge clk) begin
    nipe_q <= nipe;
    if(reset)
	pc <= 9'd0;
    else if(nipe)
	pc <= nip;
    else if(adv)
	pc <= npc;
    if(adv_q)
	mr_s <= mr;
    adv_q <= adv;
    jump_q <= jump;
end

tscram ram0 (.rclk(clk), .ra(pc), .rd(mr), .wclk(wclk), .we(we), .wa(wa), .wd(wd[29:0]));

// COME FROM loop

reg [8:0] loop0start = 9'd0, loop1start = 9'd0;
reg [8:0] loop0end = 9'd0, loop1end = 9'd0;
reg [12:0] loop0cnt = 13'd0, loop1cnt = 13'd0;

wire match0 = (pc == loop0end) && (loop0cnt != 0);
wire match1 = (pc == loop1end) && (loop1cnt != 0);
assign loop = match0 || match1;
assign looppc = match0 ? loop0start : loop1start;

always @(posedge clk) begin
    if(!jump_q && adv && rr[29:28] == 2'b01) begin
	if(rr[15]) begin
	    loop1start <= pc;
	    loop1end <= rr[24:16];
	    loop1cnt <= rr[12:0];
	end else begin
	    loop0start <= pc;
	    loop0end <= rr[24:16];
	    loop0cnt <= rr[12:0];
	end
    end else begin
	if(match0 && adv)
	    loop0cnt <= loop0cnt - 13'h0001;
	else if(match1 && adv)
	    loop1cnt <= loop1cnt - 13'h0001;
    end
end

// output-and-wait

reg [19:0] sr = 20'h03E48;
wire [19:0] msr = power ? sr : 20'h00000;
wire done = msr[19];
wire p_ppivs = msr[18];
wire p_ppihs = msr[17];
wire p_clk_yl = msr[16];
wire p_clk_y = msr[15];
wire p_clk_yr = msr[14];
wire p_sync_yl = msr[13];
wire p_sync_y = msr[12];
wire p_sync_yr = msr[11];
wire p_pc = msr[10];
wire p_shs = msr[9];
wire p_select = msr[8];
wire p_reset = msr[7];
wire p_shr = msr[6];
wire p_syl = msr[5];
wire p_syr = msr[4];
wire p_sync_x = msr[3];
wire p_clk_x = msr[2];
wire p_rsmp = msr[1];
wire p_vsmp = msr[0];

reg [7:0] dlycnt = 8'd0;
wire dlyadv = (dlycnt == 8'd0);
always @(posedge clk) begin
    if(!jump_q && adv && rr[29:28] == 2'b00) begin
	dlycnt <= rr[27:20];
	sr <= rr[19:0];
    end else if(dlycnt)
	dlycnt <= dlycnt - 8'h01;
end

// sync-to-source

reg [1:0] syncwait = 2'b00;
wire syncadv = (syncwait == 2'b00);
reg mclk_q, sync_q;
always @(posedge clk) begin
    if(!jump_q && adv && rr[29:28] == 2'b11) begin
	syncwait[0] <= !rr[0];
	syncwait[1] <= rr[0];
    end else begin
	if(mclk && !mclk_q)
	    syncwait[0] <= 1'b0;
	if(sync && !sync_q)
	    syncwait[1] <= 1'b0;
    end
    mclk_q <= mclk;
    sync_q <= sync;
end

assign adv = dlyadv && syncadv;

// generate ppiclk gating signal

reg is_vsmp_cycle;
reg p_vsmp_q;
reg [8:0] mclk_vsmp_dly;
assign ppi_clk_gate = |(mclk_vsmp_dly[8:7]);
always @(posedge clk) begin
    p_vsmp_q <= p_vsmp;

    if(mclk && !mclk_q) begin
	is_vsmp_cycle <= 1'b0;
	mclk_vsmp_dly[0] <= is_vsmp_cycle;
	mclk_vsmp_dly[8:1] <= mclk_vsmp_dly[7:0];
    end

    if(p_vsmp_q && !p_vsmp)
	is_vsmp_cycle <= 1'b1;
end

// register all outputs on IOBs for consistent timing

(* IOB = "TRUE" *) FD fd_select (.C(clk), .D(p_select), .Q(t_select));
(* IOB = "TRUE" *) FD fd_clk_yl (.C(clk), .D(p_clk_yl), .Q(t_clk_yl));
(* IOB = "TRUE" *) FD fd_reset (.C(clk), .D(p_reset), .Q(t_reset));
(* IOB = "TRUE" *) FD fd_syl (.C(clk), .D(p_syl), .Q(t_syl));
(* IOB = "TRUE" *) FD fd_sync_yl (.C(clk), .D(p_sync_yl), .Q(t_sync_yl));
(* IOB = "TRUE" *) FD fd_sync_yr (.C(clk), .D(p_sync_yr), .Q(t_sync_yr));
(* IOB = "TRUE" *) FD fd_clk_y (.C(clk), .D(p_clk_y), .Q(t_clk_y));
(* IOB = "TRUE" *) FD fd_syr (.C(clk), .D(p_syr), .Q(t_syr));
(* IOB = "TRUE" *) FD fd_sync_y (.C(clk), .D(p_sync_y), .Q(t_sync_y));
(* IOB = "TRUE" *) FD fd_clk_yr (.C(clk), .D(p_clk_yr), .Q(t_clk_yr));
(* IOB = "TRUE" *) FD fd_sync_x (.C(clk), .D(p_sync_x), .Q(t_sync_x));
(* IOB = "TRUE" *) FD fd_clk_x (.C(clk), .D(p_clk_x), .Q(t_clk_x));
(* IOB = "TRUE" *) FD fd_pc (.C(clk), .D(p_pc), .Q(t_pc));
(* IOB = "TRUE" *) FD fd_shr (.C(clk), .D(p_shr), .Q(t_shr));
(* IOB = "TRUE" *) FD fd_shs (.C(clk), .D(p_shs), .Q(t_shs));

(* IOB = "TRUE" *) FD fd_mclk (.C(clk), .D(p_mclk), .Q(t_mclk));
(* IOB = "TRUE" *) FD fd_rsmp (.C(clk), .D(p_rsmp), .Q(t_rsmp));
(* IOB = "TRUE" *) FD fd_vsmp (.C(clk), .D(p_vsmp), .Q(t_vsmp));

(* IOB = "TRUE" *) FD fd_ppiclk (.C(clk), .D(p_ppiclk), .Q(t_ppiclk));
(* IOB = "TRUE" *) FD fd_ppihs (.C(clk), .D(p_ppihs), .Q(t_ppihs));
(* IOB = "TRUE" *) FD fd_ppivs (.C(clk), .D(p_ppivs), .Q(t_ppivs));

endmodule
