`define WR_TYPE_ASYNC	2'b00
`define WR_TYPE_SDRAM   2'b01
`define WR_TYPE_RDATA   2'b11

module txmgr(clk0, mrst,
    wr_vld, wr_type, wr_addr, wr_data,
    sdreq_vld, sdreq_count, sdreq_addr,
    rd_vld, rd_addr, rd_dvld, rd_data,
    tx_dvld, tx_data,
    tx_pvld, tx_plen, tx_pflg,
    cr_vld, cr_count,
    irq_free);

input         clk0;
input         mrst;

input         wr_vld;
input   [1:0] wr_type;
input  [23:0] wr_addr;
input  [31:0] wr_data;

output        sdreq_vld;
output [11:0] sdreq_count;
output [23:0] sdreq_addr;

input         rd_vld;
input  [14:0] rd_addr;
output        rd_dvld;
output [31:0] rd_data;

output        tx_dvld;
output [31:0] tx_data;

output        tx_pvld;
output [13:0] tx_plen;
output  [1:0] tx_pflg;

input         cr_vld;
input  [11:0] cr_count;

output        irq_free;

wire          wr_vld_mirr   = wr_vld && (((wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b00)) ||
                                         ((wr_type == `WR_TYPE_SDRAM) && (wr_addr[23:2]  == 22'h00)));
wire          wr_vld_start  = wr_vld_mirr && (wr_addr[1:0] == 2'b00);
wire          wr_vld_len    = wr_vld_mirr && (wr_addr[1:0] == 2'b01);
wire          wr_vld_commit = wr_vld_mirr && (wr_addr[1:0] == 2'b10);

reg    [11:0] dfifo_cred;

wire   [13:0] wr_round_4 = wr_data[13:0] - 14'h0001;
wire   [11:0] wr_words_m1 = wr_round_4[13:2];

wire          wr_len_ok = (wr_words_m1 < dfifo_cred);

reg           tx_cap_dfifo;
reg           tx_cap_mrfifo;
reg    [23:0] tx_start;
reg    [23:0] tx_end;
reg    [13:0] tx_len;
reg     [1:0] tx_flags;

wire          tx_final_word = wr_vld && (wr_type == `WR_TYPE_SDRAM) && (wr_addr == tx_end);
wire          tx_final_dfifo = tx_final_word && tx_cap_dfifo;
wire          tx_final_mrfifo = (tx_final_word && tx_cap_mrfifo) || wr_vld_commit;
reg           tx_final_mrfifo_q;

wire          wr_hit_range = tx_cap_dfifo && (wr_addr >= tx_start) && (wr_addr <= tx_end);

reg           tx_dvld;
reg    [31:0] tx_data;

reg           tx_pvld;
reg    [13:0] tx_plen;
reg     [1:0] tx_pflg;
reg    [23:0] tx_pstart;

wire          mrf_rv;
wire          mrf_re = wr_vld && (wr_type == `WR_TYPE_RDATA) && wr_addr[23];
wire   [23:0] mrf_start;
wire   [13:0] mrf_len;
wire    [1:0] mrf_flags;

wire   [13:0] mrf_round_4 = mrf_len + 14'h0003;
wire   [11:0] mrf_words = mrf_round_4[13:2];
wire          mrf_len_ok = (mrf_words <= dfifo_cred);

reg           mrf_busy;

reg           sdreq_vld;
reg    [11:0] sdreq_count;
reg    [23:0] sdreq_addr;

wire   [11:0] dfifo_cred_p = cr_vld ? (dfifo_cred + cr_count) : dfifo_cred;

always @(posedge clk0) begin
    if(wr_vld_start)
	tx_start <= wr_data[25:2];

    if(wr_vld_len) begin
	tx_len   <= wr_data[13:0];
	tx_flags <= wr_data[31:30];
	tx_end   <= tx_start + { 10'h000, wr_words_m1 };
	tx_cap_dfifo  <=  wr_len_ok && !mrf_rv;
	tx_cap_mrfifo <= !wr_len_ok ||  mrf_rv;
    end

    tx_final_mrfifo_q <= tx_final_mrfifo;

    tx_dvld <= 1'b0;
    if(wr_vld && (((wr_type == `WR_TYPE_SDRAM) && wr_hit_range) || (wr_type == `WR_TYPE_RDATA))) begin
	tx_dvld <= 1'b1;
	tx_data <= wr_data;
	dfifo_cred <= dfifo_cred_p - 12'h001;
    end else
        dfifo_cred <= dfifo_cred_p;

    tx_pvld <= 1'b0;
    if(tx_final_dfifo) begin
	tx_pvld      <= 1'b1;
	tx_plen      <= tx_len;
	tx_pflg      <= tx_flags;
	tx_pstart    <= tx_start;
	tx_cap_dfifo <= 1'b0;
    end

    if(tx_final_mrfifo) begin
	tx_cap_mrfifo <= 1'b0;
	tx_cap_dfifo  <= 1'b0;
    end

    sdreq_vld <= 1'b0;
    if(mrf_rv && !mrf_busy && mrf_len_ok) begin
	mrf_busy    <= 1'b1;
	sdreq_vld   <= 1'b1;
	sdreq_count <= mrf_words;
	sdreq_addr  <= mrf_start;
    end

    if(mrf_re) begin
	mrf_busy  <= 1'b0;
	tx_pvld   <= 1'b1;
	tx_plen   <= mrf_len;
	tx_pflg   <= mrf_flags;
	tx_pstart <= mrf_start;
    end

    if(mrst) begin
	dfifo_cred <= 12'h3FF;
	mrf_busy <= 1'b0;
    end
end

xfifo256x40 mrfifo (.clk(clk0), .rst(mrst), .we(tx_final_mrfifo_q), .wd({tx_start, tx_len, tx_flags}), .re(mrf_re), .rv(mrf_rv), .rd({mrf_start, mrf_len, mrf_flags}));

wire          ff_re;
wire          ff_rv;
wire   [23:0] ff_rd;

xfifo256x24 ffifo (.clk(clk0), .rst(mrst), .we(tx_pvld), .wd(tx_pstart), .re(ff_re), .rv(ff_rv), .rd(ff_rd));

assign        ff_re   = rd_vld && (rd_addr[14:11] == 4'b0000) && (rd_addr[1:0] == 2'b11);
wire          rd_dvld = ff_re;
wire   [31:0] rd_data = ff_rv ? { 6'h00, ff_rd, 2'b00 } : 32'hFFFFFFFF;

wire          irq_free = ff_rv;

endmodule
