#define DEBUG 1

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/ethernet.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <linux/sockios.h>

#define LINK_EPROTO     0x6342

#define LINK_REQ_ENUM	0x0100
#define LINK_REQ_SYNC   0x0101
#define LINK_REQ_WBUF   0x2000
#define LINK_REQ_RBUF   0x2001
#define LINK_REQ_WBDF   0x2002
#define LINK_REQ_WBFF   0x2003
#define LINK_REQ_SONN   0x3000
#define LINK_REQ_SOFF   0x3001
#define LINK_REQ_SCAP   0x3002

#define LINK_REPLY      0x8000

#define LINK_SIZE_SHORT 24
#define LINK_SIZE_LONG  280
struct link_pkt {
    unsigned char  dst[6];	/* +0 */
    unsigned char  src[6];	/* +6 */
    unsigned short eproto;	/* +12 */
    unsigned char  seq;		/* +14 */
    unsigned char  ack;		/* +15 */
    unsigned short req;		/* +16 */
    unsigned short _rsvd;	/* +18 */
    unsigned int   param;	/* +20 */
    unsigned char  data[256];	/* +24 */
};

int eth_skt;
unsigned char eth_hwaddr[6];

int eth_open(char *netif)
{
    struct ifreq ifr;
    int ifindex;
    struct sockaddr_ll saddr;

    eth_skt = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
    if(eth_skt == -1) {
	perror("Opening raw ethernet socket failed");
	return 1;
    }

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, netif, sizeof(ifr.ifr_name)-1);

    if(ioctl(eth_skt, SIOCGIFINDEX, &ifr) == -1) {
	perror("Getting interface index failed");
	close(eth_skt);
	return 1;
    }
    ifindex = ifr.ifr_ifindex;

    if(ioctl(eth_skt, SIOCGIFHWADDR, &ifr) == -1) {
	perror("Getting interface MAC address failed");
	close(eth_skt);
	return 1;
    }
    memcpy(eth_hwaddr, ifr.ifr_hwaddr.sa_data, sizeof(eth_hwaddr));

#if DEBUG
    fprintf(stderr, "Using %s (index %d), %02X:%02X:%02X:%02X:%02X:%02X.\n", netif, ifindex,
	eth_hwaddr[0], eth_hwaddr[1], eth_hwaddr[2], eth_hwaddr[3], eth_hwaddr[4], eth_hwaddr[5]);
#endif

    memset(&saddr, 0, sizeof(saddr));
    saddr.sll_family   = AF_PACKET;
    saddr.sll_protocol = htons(ETH_P_ALL);
    saddr.sll_ifindex  = ifindex;
    if(bind(eth_skt, (struct sockaddr *)&saddr, sizeof(saddr)) == -1) {
	perror("Binding socket to interface failed");
	close(eth_skt);
	return 1;
    }

    return 0;
}

int eth_send(void *buf, unsigned len)
{
    memcpy(((char *)buf)+6, eth_hwaddr, 6);

    if(send(eth_skt, buf, len, 0) != len) {
	perror("Failed to send packet in one piece");
	return 1;
    }

    return 0;
}

int eth_waitrecv(void *buf, unsigned len, unsigned *rxlen, int timeout)
{
    struct pollfd pfd;
    int ret;

    pfd.fd = eth_skt;
    pfd.events = POLLIN;
    ret = poll(&pfd, 1, timeout);
    if(ret < 0 || (pfd.revents & (POLLERR | POLLHUP | POLLNVAL))) {
	perror("Poll on socket failed");
	return 1;
    }

    if(!ret) {
	if(rxlen)
	    *rxlen = 0;
	return 0;
    }

    ret = recv(eth_skt, buf, len, 0);
    if(ret < 0) {
	perror("Receive failed");
	return 1;
    }

    if(rxlen)
	*rxlen = ret;
    return 0;
}

char *ethif;
char *filename;
int loaddsp;
int pingonly;

int parseargs(int argc, char *argv[])
{
    int i, fn_exist = 0;

    ethif = strdup("eth1");
    filename = strdup("bigcam.bin");
    pingonly = 0;
    loaddsp = 0;

    for(i=1; i<argc; i++) {
	if(!strcmp(argv[i], "-i") && i+1 < argc) {
	    free(ethif);
	    ethif = strdup(argv[i+1]);
	    i ++;
	    continue;
	}

	if(!strcmp(argv[i], "-p")) {
	    pingonly = 1;
	    continue;
	}

	if(!strcmp(argv[i], "-d") && i+1 < argc && !fn_exist) {
	    free(filename);
	    filename = strdup(argv[i+1]);
	    loaddsp = 1;
	    fn_exist = 1;
	    i ++;
	    continue;
	}

	if(!strcmp(argv[i], "-f") && i+1 < argc && !fn_exist) {
	    free(filename);
	    filename = strdup(argv[i+1]);
	    loaddsp = 0;
	    fn_exist = 1;
	    i ++;
	    continue;
	}

	printf("usage: bc_prog [-i <net_if>] [-d <dsp.ldr> | -f <fpga.bin> | -p]\n");
	return 1;
    }

    return 0;
}

#define STATE_ENUM	0
#define STATE_SYNC      1
#define STATE_CONN      2

int main(int argc, char *argv[])
{
    struct link_pkt txpkt, rxpkt;
    unsigned txlen, rxlen;
    int state, retry;
    FILE *f;

    srandom(time(NULL));

    if(parseargs(argc, argv))
	return 1;

    if(eth_open(ethif))
	return 1;

    memset(&txpkt, 0, sizeof(txpkt));
    memset(txpkt.dst, 0xFF, 6);
    txpkt.eproto = LINK_EPROTO;
    txpkt.req = LINK_REQ_ENUM;
    if(eth_send(&txpkt, txlen = LINK_SIZE_SHORT))
	return 1;

    state = STATE_ENUM;
    retry = 3;

    while(1) {
	if(eth_waitrecv(&rxpkt, sizeof(rxpkt), &rxlen, 1000))
	    return 1;

	if(rxlen == 0) {
	    if(retry) {
		eth_send(&txpkt, txlen);
		retry --;
		continue;
	    }

	    fprintf(stderr, state == STATE_ENUM ? "No cameras found.\n" : "Connection lost.\n");
	    break;
	}

	if(rxlen < LINK_SIZE_SHORT || rxpkt.eproto != LINK_EPROTO)
	    continue;

	if(state == STATE_ENUM) {
	    if(rxpkt.req == (LINK_REQ_ENUM|LINK_REPLY)) {
#if DEBUG
		fprintf(stderr, "Found camera %02X:%02X:%02X:%02X:%02X:%02X.\n",
		    rxpkt.src[0], rxpkt.src[1], rxpkt.src[2], rxpkt.src[3], rxpkt.src[4], rxpkt.src[5]);
#endif
	        memset(&txpkt, 0, sizeof(txpkt));
	        memcpy(txpkt.dst, rxpkt.src, 6);
	        txpkt.eproto = LINK_EPROTO;
	        txpkt.req = LINK_REQ_SYNC;
		txpkt.seq = random() & 0xFFFF;
	        if(eth_send(&txpkt, txlen = LINK_SIZE_SHORT))
		    return 1;

		state = STATE_SYNC;
		retry = 3;
	    }
	    continue;
	}

	if(state == STATE_SYNC) {
	    if(rxpkt.req == (LINK_REQ_SYNC|LINK_REPLY) && !memcmp(rxpkt.src, txpkt.dst, 6)) {
		fprintf(stderr, "Connected to %02X:%02X:%02X:%02X:%02X:%02X.\n",
		    rxpkt.src[0], rxpkt.src[1], rxpkt.src[2], rxpkt.src[3], rxpkt.src[4], rxpkt.src[5]);
	        memset(&txpkt, 0, sizeof(txpkt));
	        memcpy(txpkt.dst, rxpkt.src, 6);
	        txpkt.eproto = LINK_EPROTO;
		txpkt.seq = rxpkt.ack + 1;
		txpkt.ack = rxpkt.seq;

		state = STATE_CONN;
		retry = 3;
		break;
	    }
	    continue;
	}
    }

    if(state != STATE_CONN) {
	close(eth_skt);
	return 1;
    }

    if(pingonly) {
	close(eth_skt);
	return 0;
    }

    f = fopen(filename, "rb");
    if(!f) {
	fprintf(stderr, "File not found\n");
	return 1;
    }

    txpkt.req = LINK_REQ_WBUF;
    txpkt.param = 0;
    while(fread(txpkt.data, 1, 256, f) > 0) {
#if DEBUG
	fprintf(stderr, "Transmitting page at 0x%08X...\n", txpkt.param);
#endif
        if(eth_send(&txpkt, LINK_SIZE_LONG))
	    return 1;

	for(retry = 0; retry < 4; retry ++) {
	    if(eth_waitrecv(&rxpkt, sizeof(rxpkt), &rxlen, 1000))
		return 1;
	    if(rxlen && memcmp(rxpkt.src, txpkt.dst, 6)) {
		retry ++;
		continue;
	    }

	    if(rxlen == 0 || rxpkt.ack != txpkt.seq || rxpkt.req != (txpkt.req^LINK_REPLY)) {
	        if(eth_send(&txpkt, LINK_SIZE_LONG))
		    return 1;
		continue;
	    }
	    break;
	}
	if(retry >= 3) {
	    fprintf(stderr, "Connection lost.\n");
	    return 1;
	}

	txpkt.param += 256;
	txpkt.seq ++;
	txpkt.ack = rxpkt.seq;
	memset(txpkt.data, 0, 256);
    }
    fclose(f);

#if DEBUG
    fprintf(stderr, "Sending %s Flash Write command for 0x%08X.\n", loaddsp ? "DSP" : "FPGA", txpkt.param);
#endif
    txpkt.req = loaddsp ? LINK_REQ_WBDF : LINK_REQ_WBFF;
    if(eth_send(&txpkt, LINK_SIZE_LONG))
	return 1;
    for(retry = 0; retry < 4; retry ++) {
	if(eth_waitrecv(&rxpkt, sizeof(rxpkt), &rxlen, 1000))
	    return 1;
	if(rxlen && memcmp(rxpkt.src, txpkt.dst, 6)) {
	    retry ++;
	    continue;
	}

	if(rxlen == 0 || rxpkt.ack != txpkt.seq || rxpkt.req != (txpkt.req^LINK_REPLY)) {
	    if(eth_send(&txpkt, LINK_SIZE_LONG))
		return 1;
	    continue;
	}
	break;
    }
    if(retry >= 3) {
        fprintf(stderr, "Connection lost.\n");
        return 1;
    }

    close(eth_skt);
    return 0;
}
